package Logic;

import Entity.Device;
import Entity.Graph;
import Entity.Node;
import Utils.GraphUtil;

import java.util.Date;
import java.util.LinkedList;

public class Check {
    //根据所有的路建立的图
    private Graph graph;
    //需要检查的设备链表
    private LinkedList<Device> devices;

    public Graph getGraph() {
        return graph;
    }

    public void setGraph(Graph graph) {
        this.graph = graph;
    }

    public LinkedList<Device> getDevices() {
        return devices;
    }

    public void setDevices(LinkedList<Device> devices) {
        this.devices = devices;
    }

    public Check(Graph graph, LinkedList<Device> devices) {
        this.graph = graph;
        this.devices = devices;
    }

    /**
     * 检查所有的设备是否合理
     *
     * 2.检查轨迹点的合理性：首先判断查出来的轨迹点是否超过三个点，如果没有超过三个轨迹点，认为本次出行无效，直接删除
     * 其次，假设第一个轨迹点是正确的，检查下一个点S(i)和前一个点之间S(i-1)的行车速度是否正确，如果不正确，
     * 判断判断当前点和后一个点速度是否正确，如果不正确，
     * @return  如果设备的数量<3 false,否则是true
     */
    public boolean CheckAllDevices(){
        int i;
        //判断一次出行
        OneTravel();
        for(i = 0;i<devices.size()-1;i++){
            if(devices.get(i).getNodeID() == devices.get(i+1).getNodeID()){
                devices.remove(i+1);
                i--;
            }
        }
        if(devices.size() < 3){
            return false;
        }
        for(i = 1;i<devices.size()-1;i++){
            Device device1 = devices.get(i-1);
            Device device2 = devices.get(i);
            Device device3 = devices.get(i+1);
            if(!IsErrorSpeed(device1,device2)){
                if(i == 1){
                    devices.remove(device1);
                }
                else if(!IsErrorSpeed(device2,device3)){
                    devices.remove(device2);
                }
                else {
                    if(!IsLookBack(device1,device2,device3)){
                        devices.remove(device2);
                    }
                }
            }
        }
        if(!IsErrorSpeed(devices.get(devices.size()-2),devices.get(devices.size()-1))){
            devices.remove(devices.size()-1);
        }
        return true;
    }
    //找到一次出行被拍到的设备ID

    /**
     * 判断拍摄时间间隔(判断一次出行的依据):如果前后两个点之间的时间间隔超过1800秒，则认为是一次出行结束，将轨迹点串截断。
     */
    public void OneTravel(){
        int i,j;
        int size = devices.size();
        for(i = 0;i<devices.size()-1;i++){
            long data1 = devices.get(i).getShotTime().getTime();
            long data2 = devices.get(i+1).getShotTime().getTime();
            if(data2-data1 > 60*60*1000){   //设置一小时为一次出行
                for(j = 0;j<size-i-1;j++){
                    devices.remove(i+1);
                }
                break;
            }
        }
    }

    /**
     * 判断两个设备之间车的速度是否正确
     * 用最短路径求出两个设备之间的最短路径，求出长度
     * 限制速度为30m/s，即108km/h
     * @param device1   第一个设备对象
     * @param device2   第二个设备对象
     * @return  如果车速正，返回true,否则，，返回true
     */
    public boolean IsErrorSpeed(Device device1, Device device2){
        Date date1 = device1.getShotTime();
        Date date2 = device2.getShotTime();
        Dijkstra dijkstra = new Dijkstra(graph);
        //计算速度单位为m/s
        //速度大小限制为30m/s
        if(dijkstra.getShortPath(device1.getNodeID(),device2.getNodeID())){
            Double leng = dijkstra.getLength();
            Double distanceTime = (((double)date2.getTime()-(double)date1.getTime()))/1000;
            if(leng == null){
                return false;
            }
            Double speed = leng/distanceTime;
            if(speed > 30){
                return false;
            }
        }else {
            return false;
        }
        return true;
    }
    /**
     * 判断三个点之间中间点是否是差错点
     * @param device1   第一个设备对象
     * @param device2   第二个设备对象
     * @param device3   第三个设备对象
     * 当前点到前后两个点之间的距离的最小值比前后两个点之间距离值的2倍还要大，则认为当前点是不正确点
     * @return  是错误点，值为false，否则，值为true
     */
    public boolean IsLookBack(Device device1,Device device2,Device device3){
        Dijkstra dijkstra1 = new Dijkstra(graph);
        Dijkstra dijkstra2 = new Dijkstra(graph);
        Dijkstra dijkstra3 = new Dijkstra(graph);
        LinkedList<Node> nodes1 = new LinkedList<>();
        LinkedList<Node> nodes2 = new LinkedList<>();
        LinkedList<Node> nodes3 = new LinkedList<>();
        //通过dijstra算法求得两个点之间的最短路径
        if(dijkstra1.getShortPath(device1.getNodeID(),device2.getNodeID())){
            nodes1 = dijkstra1.getRoadPath();
        }
        if(dijkstra1.getShortPath(device2.getNodeID(),device3.getNodeID())){
            nodes2 = dijkstra2.getRoadPath();
        }
        if(dijkstra1.getShortPath(device1.getNodeID(),device3.getNodeID())){
            nodes3 = dijkstra3.getRoadPath();
        }
        double leng1 = getPathLeng(nodes1);
        double leng2 = getPathLeng(nodes2);
        double leng3 = getPathLeng(nodes3);
        //如果当前点的前一个点后后一个点的距离大于2倍的前一个点到后一个点的距离，则认为当前点是差错点
        if(leng1+leng2 > 2*leng3){
            return false;
        }
        return true;
    }
    /**
     * 给定一条路上的点，求出这段路的长度
     * @param nodes 这条路经过路网上的点
     * @return  路的长度
     */
    public double getPathLeng(LinkedList<Node> nodes){
        double totalLeng = 0;
        int id1,id2;
        for(int i = 0;i<nodes.size()-1;i++){
            id1 = nodes.get(i).getNodeID();
            id2 = nodes.get(i+1).getNodeID();
            totalLeng += GraphUtil.findLeng(graph,id1,id2);
        }
        return totalLeng;
    }
}
